unit ICLinkInSubs;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all. *)

(* This file contains a bunch of subroutines that are used by the link-in implementation
	(contained in the glue file or the component).  These routines may also be called
	by the component code itself.
	
	The bulk of the routines in this module fall into the following categories:
	
	o environment checking
	o wacky file system stuff
	o AppleEvent stuff
	
	Note that all routines exported by this module have the prefix "ICU".  This
	is because this module is compiled into "ICGlue.o", which is linked with
	other applications.  Without the suffix, common routines like FSpGetCatInfo
	might clash with the same routine in the client application, with possibly
	very weird results.
*)

interface

	uses
		Types,
		Files,
		QuickDraw,
		Aliases, 

		ICTypes,
		ICKeys;
		
	(* ***** Environment Checking Routines ***** *)

	(* The folowing four routines do simple Gestalt checking.  Note that
		IC relies on the environment's Gestalt glue to ensure that Gestalt
		is safe to call with any system version.
	*)

	function ICUHaveFindFolder : Boolean;
	function ICUHaveProcessManager: Boolean;
	function ICUHaveNewStandardFile: Boolean;
	function ICUHaveAliasManager: Boolean;

	function ICUCanInteract: ICError;
		(* Returns noErr if user interaction is possible.  Basically a wrapper
			around AEInteractWithUser that makes it System 6 safe.
		*)

	(* ***** File System Routines ***** *)
	
	function ICUFindFolder (vRefNum: integer; folderType: OSType; createFolder: Boolean; var foundVRefNum: integer; var foundDirID: longint): OSErr;
		inline
			$7000,				(* move.q	#0,d0		*)
			$A823;				(* _AliasDispatch	*)
		(* This is direct glue to call FindFolder.  We use this rather than calling
			FindFolder directly because all of our calls to FindFolder are guarded
			by checks for ICUHaveFindFolder, so we don't need to roll the glue
			into our code.
		*)

	function ICUFSpGetCatInfo (var fss: FSSpec; ioFDirIndex: integer; var cpb: CInfoPBRec): OSErr;
	function ICUFSpSetCatInfo (var fss: FSSpec; var cpb: CInfoPBRec): OSErr;
		(* High-level wrappers to PBGetCatInfo, PBSetCatInfo. *)

	function ICUFSSpecToFullPath (fss: FSSpec; var path: Str255): OSErr;
		(* Returns a full path for the given FSSpec.  Actually it's an approximation
			(it doesn't handle paths longer than 256 characters) but, seeing as
			IC only uses this path for display purposes, that's not a problem.
		*)

	function ICFileSpecToFSSpec (fileSpec: ICFileSpecHandle; canInteract: Boolean; var fss: FSSpec): ICError;
	function FSSpecToICFileSpec (var fss: FSSpec; fileSpec: ICFileSpecHandle): OSErr;
		(* These routines convert ICFileSpecHandles to FSSpecs and vice versa.
			An ICFileSpecHandle is basically an AliasHandle with some fields tacked
			on the front to make it usable by System 6 clients.  Under System 7 these
			routines are basically wrappers for standard Alias Manager routines.
			Note that in poth of these routines, the caller is responsible for allocating
			and deallocating the ICFileSpecHandle, and we just resize it if necessary.
		*)

	(* ***** AppleEvent Routines ***** *)
	
	function ICULaunchURL (helperCreator: OSType; urlH: Handle): ICError;
	function ICUEditPreferences (key: Str255; prefsFileSpec: FSSpec): ICError;
		(* These routines are the high level wrappers to code that
			sends the two important AppleEvents that IC deals with, namely
			the GURL event (to ask helpers to get a URL) and the Edit
			Preferences event (to ask the IC application to show
			a certain preference panel.
		*)

implementation

	uses
		GestaltEqu,
		Errors, 
		ToolUtils, 
		Packages, 
		Processes,
		AppleEvents;

	(* ***** Environment Checking Routines ***** *)

	function ICUHaveFindFolder : Boolean;
		var
			gestaltResponse : longint;
	begin
		ICUHaveFindFolder := (Gestalt(gestaltFindFolderAttr, gestaltResponse) = noErr) & 
													btst(gestaltResponse, gestaltFindFolderPresent);
	end; (* ICUHaveFindFolder *)

	function ICUHaveProcessManager : Boolean;
		var
			gestaltResponse : longint;
	begin
		ICUHaveProcessManager := (Gestalt(gestaltOSAttr, gestaltResponse) = noErr) & 
													btst(gestaltResponse, gestaltLaunchControl);
	end; (* ICUHaveProcessManager *)

	function ICUHaveNewStandardFile: Boolean;
		var
			gestaltResponse : longInt;
	begin
		ICUHaveNewStandardFile := (Gestalt(gestaltStandardFileAttr, gestaltResponse) = noErr) &
													btst(gestaltResponse, gestaltStandardFile58);
	end; (* ICUHaveNewStandardFile *)

	function ICUHaveAliasManager: Boolean;
		var
			gestaltResponse : longInt;
	begin
		ICUHaveAliasManager := (Gestalt(gestaltAliasMgrAttr, gestaltResponse) = noErr) &
													btst(gestaltResponse, gestaltAliasMgrPresent);
	end; (* ICUHaveAliasManager *)

	function ICUCanInteract: ICError;
		(* See comment in interface part. *)
		var
			err: ICError;
			gestaltResponse : longint;
	begin
		err := noErr;
		if (Gestalt(gestaltAppleEventsAttr, gestaltResponse) = noErr) &
									btst(gestaltResponse, gestaltAppleEventsPresent) then begin
			err := AEInteractWithUser(kAEDefaultTimeout, nil, nil);
		end; (* if *)
		ICUCanInteract := err;
	end; (* ICUCanInteract *)

	(* ***** File System Routines ***** *)

	function ICUFSpGetCatInfo (var fss: FSSpec; ioFDirIndex: integer; var cpb: CInfoPBRec): OSErr;
		(* High-level wrapper to PBGetCatInfo. *)
	begin
		cpb.ioVRefNum := fss.vRefNum;
		cpb.ioDirID := fss.parID;
		cpb.ioNamePtr := @fss.name;
		cpb.ioFDirIndex := ioFDirIndex;
		ICUFSpGetCatInfo := PBGetCatInfoSync(@cpb);
	end; (* ICUFSpGetCatInfo *)

	function ICUFSpSetCatInfo (var fss: FSSpec; var cpb: CInfoPBRec): OSErr;
		(* High-level wrapper to PBSetCatInfo. *)
	begin
		cpb.ioVRefNum := fss.vRefNum;
		cpb.ioDirID := fss.parID;
		cpb.ioNamePtr := @fss.name;
		ICUFSpSetCatInfo := PBSetCatInfoSync(@cpb);
	end; (* ICUFSpSetCatInfo *)

	function ICUFSSpecToFullPath (fss: FSSpec; var path: Str255): OSErr;
		(* See comment in interface part. *)
		var
			err: OSErr;
			pb: CInfoPBRec;
	begin
		if fss.parID = 1 then begin
			(* It's a volume, just return the name:. *)
			path := concat(fss.name, ':');
		end else begin
		
			(* It's a file or folder, start by putting the name at the end of the path
				and then iterate up the directory hierarchy adding directory names to
				the front of the path.
			*)
			path := fss.name;
			while (err = noErr) & (fss.parID <> 1) do begin
				err := ICUFSpGetCatInfo(fss, -1, pb);
				path := concat(fss.name, ':', path);
				fss.parID := pb.ioFlParID;
			end; (* while *)
			
		end; (* if *)
		ICUFSSPecToFullPath := err;
	end; (* ICUFSSPecToFullPath *)

	function FindVolumeByNameAndDate (name: Str31; creationDate: longInt; var vRefNum: integer): OSErr;
		(* Attempts to find a volume based on it's name and creation date.  This
			is the tricky part of our "poor man's alias resolution" scheme.  An
			ICFileSpec stores the volume name and creation date of the item
			it points to.  These are used to try to find the matching volume
			in systems that don't have the Alias Manager.  This routine implements
			that finding code.
			
			The routine takes a two phase approach.  In the first phase, it searches for
			volumes by name and creation date.  If it can't find a match, it proceeds
			to the second phase where a matching name is considered good enough.
			
			Please don't blame me for the number of "leave"s in this code.  Peter
			wrote it, and I've learnt through hard experience that I'm too stupid
			to mess with his code too much.
		*)
		var
			err: OSErr;
			phase : (kMatchNameAndCreationDate, kMatchOnlyName);
			volumeName: Str255;
			volumeIndex: integer;
			pb: HParamBlockRec;
	begin
		for phase := kMatchNameAndCreationDate to kMatchOnlyName do begin

			volumeIndex := 1;
			while true do begin
				(* Get info an the volumeIndex'th volume. *)
				volumeName := '';
				pb.ioNamePtr := @volumeName;
				pb.ioVolIndex := volumeIndex;
				err := PBGetVInfoSync(@pb);
				if err <> noErr then begin
					leave;
				end; (* if *)

				(* Check for a match. *)
				if EqualString(name, volumeName, false, true) then begin
					if (phase = kMatchOnlyName) or (pb.ioVCrDate = creationDate) then begin
						leave;
					end; (* if *)
				end; (* if *)

				volumeIndex := volumeIndex + 1;
			end; (* while *)

			(* Leave if we found a match. *)
			if err = noErr then begin
				leave;
			end; (* if *)
		end; (* for *)
		
		(* Return the vRefNum of the found volume. *)
		if err = noErr then begin
			vRefNum := pb.ioVRefNum;
		end; (* if *)
		FindVolumeByNameAndDate := err;
	end; (* FindVolumeByNameAndDate *)

	function ICFileSpecToFSSpec (fileSpec: ICFileSpecHandle; canInteract: Boolean; var fss: FSSpec): ICError;
		(* See comment in interface part. *)
		var
			err: ICError;
			junkLong: longInt;
			aliasH: AliasHandle;
			aliasCount: integer;
			aliasMatchRules: longInt;
			junkBool: Boolean;
			cpb: CInfoPBRec;
	begin
		err := noErr;
		if (err = noErr) & (GetHandleSize(Handle(fileSpec)) < sizeof(ICFileSpec)) then begin
			err := paramErr;
		end; (* if *)
		
		if err = noErr then begin

			(* Try to find it using the alias embedded in the ICFileSpec. *)
			err := -1;
			if ICUHaveAliasManager & (fileSpec^^.alias.aliasSize <> 0) then begin
				(* Make a copy of the ICFileSpecHandle. *)
				aliasH := AliasHandle(fileSpec);
				err := HandToHand(Handle(aliasH));
				if err = noErr then begin
					(* Use Munger to delete our fields from the front of the copy, thereby
						turning it into a real AliasHandle.
					*)
					junkLong := Munger(Handle(aliasH), 0, nil, sizeof(ICFileSpec) - sizeof(AliasRecord), @junkLong, 0);

					(* Call the Alias Manager to find the match. *)
					aliasCount := 1;
					aliasMatchRules := kARMSearch + kARMMountVol;
					if canInteract & (ICUCanInteract <> noErr) then begin
						aliasMatchRules := aliasMatchRules + kARMNoUI;
					end; (* if *)
					err := MatchAlias(nil, aliasMatchRules, aliasH, aliasCount, @fss, junkBool, nil, nil);

					(* Dispose our copy of the alias. *)
					DisposeHandle(Handle(aliasH));
				end; (* if *)
			end; (* if *)

			(* If it we didn't find it, try using our poor man's alias. *)
			if err <> noErr then begin

				(* Attempt to find a matching volume. *)
				err := FindVolumeByNameAndDate(fileSpec^^.vol_name, fileSpec^^.vol_creation_date, fss.vRefNum);

				(* If it worked, build an FSSpec for the item and confirm it's existance using
					GetCatInfo.
				*)
				if err = noErr then begin
					fss.parID := fileSpec^^.fss.parID;
					fss.name := fileSpec^^.fss.name;
					err := ICUFSpGetCatInfo(fss, 0, cpb);
				end; (* if *)
			end; (* if *)
		end; (* if *)

		ICFileSpecToFSSpec := err;
	end; (* ICFileSpecToFSSpec *)

	function FSSpecToICFileSpec (var fss: FSSpec; fileSpec: ICFileSpecHandle): OSErr;
		(* See comment in interface part. *)
		var
			err: OSErr;
			pb: HParamBlockRec;
			volumeName: Str63;
			aliasH: AliasHandle;
			junkLong: longInt;
	begin
		(* First resize the handle to the basic size and fill in our poor man's alias
			information.
		*)
		SetHandleSize(Handle(fileSpec), sizeof(ICFileSpec));
		err := MemError;
		if err = noErr then begin
			
			(* Get the volume information. *)
			volumeName := '';
			pb.ioNamePtr := @volumeName;
			pb.ioVRefNum := fss.vRefNum;
			pb.ioVolIndex := 0;
			err := PBGetVInfoSync(@pb);
			
			(* Fill in the basic fields of the ICFileSpec. *)
			if err = noErr then begin
				fileSpec^^.vol_creation_date := pb.ioVCrDate;
				fileSpec^^.vol_name := volumeName;
				fileSpec^^.fss := fss;
				fileSpec^^.alias.userType := OSType(0);
				fileSpec^^.alias.aliasSize := 0;
			end; (* if *)
		end; (* if *)

		(* Now, if we have the Alias Manager, create an alias and append it to the handle. 
			This is entirely optional, so we make sure that any errors encountered in this
			process don't make it out to the client.
		*)
		if (err = noErr) & ICUHaveAliasManager then begin
		
			(* Create the alias. *)
			err := NewAlias(nil, fss, aliasH);
			if err = noErr then begin
				
				(* Append it to the end of the fileSpec and then delete the dummy
					AliasRecord from the end of the original fileSpec.
				*)
				err := HandAndHand(Handle(aliasH), Handle(fileSpec));
				if err = noErr then begin
					junkLong := Munger(Handle(fileSpec), sizeof(ICFileSpec) - sizeof(AliasRecord), nil, sizeof(AliasRecord), @junkLong, 0);
				end; (* if *)
				DisposeHandle(Handle(aliasH));
			end; (* if *)
			err := noErr;
		end; (* if *)

		FSSpecToICFileSpec := err;
	end; (* FSSpecToICFileSpec *)

	(* ***** Utilities to Support the AppleEvent Routines ***** *)

	function FindProcess (processCreator : OSType; var process: ProcessSerialNumber): Boolean;
		(* Looks for a running process based on its creator type.  If it finds it, it returns
			true and the processes PSN.
		*)
		var
			info: ProcessInfoRec;
	begin
		FindProcess := false;
		process.highLongOfPSN := 0;
		process.lowLongOfPSN := kNoProcess;
		info.processInfoLength := sizeof(ProcessInfoRec);
		info.processName := nil;
		info.processAppSpec := nil;
		while GetNextProcess(process) = noErr do begin
			if (GetProcessInformation(process, info) = noErr) &
							(info.processSignature = processCreator) then begin
				FindProcess := true;
				leave;
			end; (* if *)
		end; (* while *)
	end; (* FindProcess *)

	function ConfirmApplication (var fss: FSSpec; creator: OSType): OSErr;
		(* This routine confirms when the given FSSpec points to
			an application with the given type.  We use this to sanity
			check the information returned by the desktop database,
			ie avoid launching things that are in the DTDB but don't
			actually exist!
		*)
		var
			err: OSErr;
			info: FInfo;
	begin
		err := HGetFInfo(fss.vRefNum, fss.parID, fss.name, info);
		if err = noErr then begin
			if (info.fdType <> 'APPL') or (info.fdCreator <> creator) then begin
				err := afpItemNotFound;
			end; (* if *)
		end; (* if *)
		ConfirmApplication := err;
	end; (* ConfirmApplication *)

	function ScanVolumeForApplication (vRefNum: integer; applCreator: OSType; var fss: FSSpec): OSErr;
		(* Scans the volume specified by vRefNum looking for the application
			with creator applCreator.  It sets up fss to point to the found application.
			
			Note that this routine doesn't just look for the first entry in the DTDB.  It
			tries to confirm whether that entry really points to an application and,
			if it doesn't, it keeps scanning through the DTDB for the volume.
		*)
		var
			err: OSErr;
			pbdt: DTPBRec;
			fileIndex: integer;
			found: Boolean;
	begin
		(* Get the path to the DTDB file. *)
		fss.name := '';
		pbdt.ioNamePtr := @fss.name;
		pbdt.ioVRefNum := vRefNum;
		err := PBDTGetPath(@pbdt);
		
		(* Now scan through the file looking for matching applications. *)
		if err = noErr then begin
			fileIndex := 1;
			found := false;
			repeat
				pbdt.ioIndex := fileIndex;
				pbdt.ioFileCreator := applCreator;
				err := PBDTGetAPPLSync(@pbdt);
				if err = noErr then begin
					(* We found a match, confirm that it exists.  Note that
						fss.name has already been setup by the PBDTGetAPPLSync
						call.
					*)
					fss.vRefNum := vRefNum;
					fss.parID := pbdt.ioAPPLParID;
					found := (ConfirmApplication(fss, applCreator) = noErr);
				end; (* if *)
				fileIndex := fileIndex + 1;
			until found or (err <> noErr);
		end; (* if *)
		ScanVolumeForApplication := err;
	end; (* ScanVolumeForApplication *)

	function FindApplication (applCreator: OSType; var fss: FSSpec): OSErr;
		(* This routine attempts to find an application someone in the file
			system with a matching creator.  It does this by scanning the
			desktop database on each available volume.
		*)

		function GetIndexedVolume (volumeIndex: integer; var vRefNum: integer): OSErr;
			(* This routine returns the vRefNum for the volumeIndex'th volume
				that's currently mounted.  It's in a separate procedure so that
				FindApplication doesn't waste stack space keeping a ParamBlockRec
				around while it calls ScanVolumeForApplication.
			*)
			var
				err: OSErr;
				pb: ParamBlockRec;
		begin
			pb.ioNamePtr := nil;
			pb.ioVRefNum := 0;
			pb.ioVolIndex := volumeIndex;
			err := PBGetVInfoSync(@pb);
			if err = noErr then begin
				vRefNum := pb.ioVRefNum;
			end; (* if *)
			GetIndexedVolume := err;
		end; (* GetIndexedVolume *)

		var
			err: OSErr;
			volumeIndex: integer;
			vRefNum: integer;
			found: Boolean;
	begin
		(* Index through all the volumes calling ScanVolumeForApplication for a match. *)
		found := false;
		volumeIndex := 1;
		repeat
			vRefNum := 0;
			err := GetIndexedVolume(volumeIndex, vRefNum);
			if err = noErr then begin
				err := ScanVolumeForApplication(vRefNum, applCreator, fss);
				if err = noErr then begin
					found := true;
				end else begin
					err := noErr;		(* Swallow error so we continue with next volume. *)
				end; (* if *)
			end; (* if *)
			volumeIndex := volumeIndex + 1;
		until found or (err <> noErr);

		(* If we didn't find a match, always return the same error code. This
			makes it easier for our clients to display meaningful error messages.
		*)
		if not found then begin
			err := afpItemNotFound;
		end; (* if *)
		FindApplication := err;
	end; (* FindApplication *)

	function PrepareLaunchParamBlock (var theEvent: AppleEvent; var launchThis: LaunchParamBlockRec; var paramsDesc: AEDesc): ICError;
		(* Prepare a LaunchParamBlock for launching an application.  The application
			to launch has already been set up in launchThis.launchAppSpec.  This routine
			is responsible for filling out the other fields in the LaunchParamBlock, including
			the launchAppParameters, which is set up from theEvent parameter.
			The routine also returns the AEDesc form of theEvent so that the caller
			can dispose it.  It's all very ugly, but it works well in practice.
		*)
		var
			err : ICError;
	begin
		err := AECoerceDesc(theEvent, typeAppParameters, paramsDesc);
		if err = noErr then begin
			HLock(handle(paramsDesc.dataHandle));
			launchThis.launchAppParameters := AppParametersPtr(paramsDesc.dataHandle^);
			launchThis.launchBlockID := extendedBlock;
			launchThis.launchEPBLength := extendedBlockLen;
			launchThis.launchFileFlags := 0;
			launchThis.launchControlFlags := launchContinue + launchNoFileFlags;
		end; (* if *)
		PrepareLaunchParamBlock := err;
	end; (* PrepareLaunchParamBlock *)

	function CreateGURLEvent (applCreator: OSType; urlH: Handle; var theEvent: AppleEvent): ICError;
		(* Create a GURL AppleEvent that targets the given application with urlH
			as the direct object.
		*)
		var
			err: ICError;
			err2: ICError;
			junk: ICError;
			targetAddress: AEDesc;
			s: signedByte;
	begin
		(* Create the event. *)
		err := AECreateDesc(typeApplSignature, @applCreator, sizeof(applCreator),
													targetAddress);
		err2 := AECreateAppleEvent('GURL', 'GURL', targetAddress, kAutoGenerateReturnID, 
													kAnyTransactionID, theEvent);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		
		(* Add urlH as the direct object parameter. *)
		s := HGetState(urlH);
		HLock(urlH);
		err2 := AEPutKeyPtr(theEvent, keyDirectObject, typeChar, urlH^, GetHandleSize(urlH));
		HSetState(urlH, s);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		
		(* Clean up. *)
		if err <> noErr then begin
			junk := AEDisposeDesc(theEvent);
		end; (* if *)
		junk := AEDisposeDesc(targetAddress);

		CreateGURLEvent := err;
	end; (* CreateGURLEvent *)

	function CreateEditPrefEvent (key: Str255; prefsFileSpec: FSSpec; var theEvent: AppleEvent): ICError;
		(* Create a Edit Preference AppleEvent that targets the given application with urlH
			as the direct object.
		*)
		const
			kICEditPreferenceEventClass = 'ICAp';
			kICEditPreferenceEvent = 'ICAp';
			keyEditPreferenceDestination = 'dest';
		(* These are the same constant names and values used by the Internet Config
			application.  One day I'll get the application and extension to actually share
			source and avoid this duplication.
		*)
		var
			err: ICError;
			err2: ICError;
			junk: ICError;
			targetAddress: AEDesc;
			icApplCreator: OSType;
	begin
		(* Create the event. *)
		icApplCreator := ICcreator;
		err := AECreateDesc(typeApplSignature, @icApplCreator, sizeof(icApplCreator), targetAddress);
		err2 := AECreateAppleEvent(kICEditPreferenceEventClass, kICEditPreferenceEvent, 
								targetAddress, kAutoGenerateReturnID, kAnyTransactionID, theEvent);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		
		(* Add key as the direct object parameter. *)
		err2 := AEPutKeyPtr(theEvent, keyDirectObject, typeChar, @key[1], length(key));
		if err = noErr then begin
			err := err2;
		end; (* if *)
		
		(* Edit the destination file parameter. *)
		err2 := AEPutKeyPtr(theEvent, keyEditPreferenceDestination, typeFSS,
											@prefsFileSpec, sizeof(prefsFileSpec));
		if err = noErr then begin
			err := err2;
		end; (* if *)

		(* Clean up. *)
		if err <> noErr then begin
			junk := AEDisposeDesc(theEvent);
		end; (* if *)
		junk := AEDisposeDesc(targetAddress);

		CreateEditPrefEvent := err;
	end; (* CreateEditPrefEvent *)

	function LaunchFSSpec (var applFileSpec: FSSpec; theEvent: AppleEvent): ICError;
		(* Launch application denoted by applFileSpec passing it the AppleEvent
			as its application parameter.
		*)
		var
			err: ICError;
			junk: ICError;
			launchThis: LaunchParamBlockRec;
			paramsDesc: AEDesc;
	begin
		(* Setup the launch parameter block. *)
		launchThis.launchAppSpec := @applFileSpec;
		err := PrepareLaunchParamBlock(theEvent, launchThis, paramsDesc);
		
		(* Launch the application. *)
		if err = noErr then begin
			err := LaunchApplication(@launchThis);
		end; (* if *)
		
		(* If it fails because we're out of memory, try launching the application in
			the minimum partition.
		*)
		if err = memFullErr then begin
			launchThis.launchControlFlags := bor(launchThis.launchControlFlags, launchUseMinimum);
			err := LaunchApplication(@launchThis);
		end; (* if *)
		
		(* Dispose the application parameters AEDesc that was created by
			PrepareLaunchParamBlock.
		*)
		junk := AEDisposeDesc(paramsDesc);

		LaunchFSSpec := err;
	end; (* LaunchFSSpec *)

	function SendEvent (theEvent: AppleEvent; applCreator: OSType): ICError;
		(* Send an AppleEvent to a given application, denoted by its
			creator code.  The implementation handles two cases: if the
			application is already running, it sends it the AppleEvent directly.
			If the application is not running, it launches it with the AppleEvent
			packaged into the application parameters.
		*)
		var
			err: ICError;
			junk: ICError;
			applPSN: ProcessSerialNumber;
			applFileSpec: FSSpec;
			reply: AppleEvent;
	begin
		if FindProcess(applCreator, applPSN) then begin
			junk := SetFrontProcess(applPSN);
			err := AESend(theEvent, reply, kAENoReply, kAEHighPriority, kNoTimeOut, nil, nil);
		end else begin
			err := FindApplication(applCreator, applFileSpec);
			if err = noErr then begin
				err := LaunchFSSpec(applFileSpec, theEvent);
			end; (* if *)
		end; (* if *)
		SendEvent := err;
	end; (* SendEvent *)

	(* ***** AppleEvent Routines ***** *)

	function ICUEditPreferences (key: Str255; prefsFileSpec: FSSpec): ICError;
		(* Implements the core of the ICEditPreferences call.  Sends
			an Edit Preferences AppleEvent to the Internet Config application,
			telling it to edit the given key and preferences file.
		*)
		var
			err: ICError;
			junk: ICError;
			theEvent: AppleEvent;
	begin
		if ICUHaveProcessManager then begin
			err := CreateEditPrefEvent(key, prefsFileSpec, theEvent);
			if err = noErr then begin
				err := SendEvent(theEvent, ICcreator);
			end; (* if *)
			junk := AEDisposeDesc(theEvent);
		end else begin
			err := unimpErr;
		end; (* if *)
		ICUEditPreferences := err;
	end; (* ICUEditPreferences *)

	function ICULaunchURL (helperCreator: OSType; urlH: Handle): ICError;
		(* Implements the core of the ICLaunchURL call.  Sends
			an GURL AppleEvent to the specified helper application,
			telling it to launch the given URL.
		*)
		var
			err: ICError;
			junk: ICError;
			theEvent: AppleEvent;
	begin
		if ICUHaveProcessManager then begin
			err := CreateGURLEvent(helperCreator, urlH, theEvent);
			if err = noErr then begin
				err := SendEvent(theEvent, helperCreator);
			end; (* if *)
			junk := AEDisposeDesc(theEvent);
		end else begin
			err := unimpErr;
		end; (* if *)
		ICULaunchURL := err;
	end; (* ICULaunchURL *)

end. (* ICLinkInSubs *)
